package com.tidecloud.dataacceptance.service;

import javax.xml.bind.DatatypeConverter;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import io.netty.buffer.ByteBuf;
import io.netty.channel.Channel;
import io.netty.channel.ChannelHandlerContext;

/**
 * 16 进制 2进制  抽象
 */
public abstract class HexBinaryAcceptanceHandlerAdapter extends AcceptanceInboundHandlerAdapter {
    private static final Logger logger = LoggerFactory.getLogger(HexBinaryAcceptanceHandlerAdapter.class);

    @Override
    public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception {
        super.channelRead(ctx, msg);
        ByteBuf in = (ByteBuf) msg;

        try {
            if (!in.isReadable()) {
                return;
            }
            printAcceptanceData(in, ctx);
            handle(in, ctx.channel());
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        } finally {
            // release每次只把引用次数减1，通过日志发现，这里是in对象最后一次被引用的地方
            // 可以推断出在handler方法中必然有把in的引用次数增加但是没有做释放的操作
            // 代码太多了，不好排查。这里先做一个兜底的操作，如果释放不成功，多次释放，直到成功为止
            // 事实证明不是这里的问题，但是放在这里也没有问题，把日志改成debug级别，打印的太多了
            logger.debug("the ref count is: {}", in.refCnt());
            boolean isReleaseSuccess = false;
            while (!isReleaseSuccess) {
                isReleaseSuccess = in.release();
                logger.debug("is release success: {}", isReleaseSuccess);
            }
        }

        ctx.flush();
    }

    protected void printAcceptanceData(ByteBuf dataByteBuf, ChannelHandlerContext ctx) {
        //ByteBuf dataByteBufCopy = dataByteBuf.copy();
        dataByteBuf.markReaderIndex();
        byte[] dataByteArray = new byte[dataByteBuf.readableBytes()];
        dataByteBuf.readBytes(dataByteArray);
        String printHexBinary = DatatypeConverter.printHexBinary(dataByteArray);
        if (channelDeviceMap.get(ctx.channel()) != null) {
            MDC.put(MDC_DEVICEID, channelDeviceMap.get(ctx.channel()));
            logger.info("设备: [{}] 传入数据为 : {}", channelDeviceMap.get(ctx.channel()), printHexBinary);
        } else {
            logger.info("设备传入数据为 : {}", printHexBinary);
        }
        dataByteBuf.resetReaderIndex();
        //dataByteBufCopy.release();
    }

    abstract protected void handle(ByteBuf in, Channel channel) throws Exception;

}
